/**
 * Copyright (c) 2022 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "DawnGLFWHost.hpp"
#include "game/DawnGame.hpp"
#include "dawnopengl.hpp"

using namespace Dawn;

GLFWwindow *window;

// GLFW Callbacks
void glfwOnError(int error, const char* description) {
  fputs(description, stderr);
}

// Host
DawnHost::DawnHost() {
  this->data = std::make_unique<DawnHostData>();
  this->data->window = nullptr;
}

int32_t DawnHost::init(std::weak_ptr<DawnGame> game) {
  // Init GLFW
  if(!glfwInit()) return DAWN_GLFW_INIT_RESULT_INIT_FAILED;

  glfwSetErrorCallback(&glfwOnError);
  
  // Setup window hints 
  glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, false);
  
  // Create Window
  window = glfwCreateWindow(
    DAWN_GLFW_WINDOW_WIDTH_DEFAULT,
    DAWN_GLFW_WINDOW_HEIGHT_DEFAULT,
    "Dawn", NULL, NULL
  );
  if(window == nullptr) {
    glfwTerminate();
    return DAWN_GLFW_INIT_RESULT_WINDOW_CREATE_FAILED;
  }

  // Load GLAD
  glfwMakeContextCurrent(window);
  glfwSwapInterval(0);
  gladLoadGLLoader((GLADloadproc)glfwGetProcAddress);

  // Initialize the game
  if(auto g = game.lock()) {
    auto result = g->init();
    if(result != DAWN_GAME_INIT_RESULT_SUCCESS) return result;
  } else {
    return DAWN_GLFW_INIT_RESULT_GAME_LOCK_FAILED;
  }

  return DAWN_HOST_INIT_RESULT_SUCCESS;
}

int32_t DawnHost::start(std::weak_ptr<DawnGame> game) {
  double_t time, newTime;
  float_t fDelta;
  int32_t updateResult;

  // Main Render Loop 
  time = 0.0f;
  while(!glfwWindowShouldClose(window)) {
    // Determine the delta.
    newTime = glfwGetTime();
    fDelta = (float_t)(newTime - time);
    time = newTime;

    // Perform update
    updateResult = this->update(game, fDelta);

    // Did the update complete successfully?
    if(updateResult == DAWN_HOST_UPDATE_RESULT_EXIT) {
      break;
    } else if(updateResult != DAWN_HOST_UPDATE_RESULT_SUCCESS) {
      return DAWN_GLFW_START_RESULT_UPDATE_FAILED;
    }
    
    // Tick the engine.
    glfwSwapBuffers(window);
    
    // Update events
    glfwPollEvents();
  }

  return DAWN_HOST_START_RESULT_EXIT_SUCCESS;
}

int32_t DawnHost::update(std::weak_ptr<DawnGame> game, float_t delta) {
  if(auto g = game.lock()) {
    auto ret = g->update(delta);
    switch(ret) {
      case DAWN_GAME_UPDATE_RESULT_SUCCESS:
        return DAWN_HOST_UPDATE_RESULT_SUCCESS;
      
      case DAWN_GAME_UPDATE_RESULT_EXIT:
        return DAWN_HOST_UPDATE_RESULT_EXIT;

      default:
        return ret;
    }
  }
}

void DawnHost::unload(std::weak_ptr<DawnGame> game) {
  glfwDestroyWindow(this->data->window);
  this->data->window = nullptr;
  glfwTerminate();
}

DawnHost::~DawnHost() {

}